; Copyright (c) 2007 Oak Circle Software, Inc.
; 
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
; 
; The above copyright notice and this permission notice shall be included in
; all copies or substantial portions of the Software.
; 
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
; THE SOFTWARE.

;;;; Created on 2007-10-10 18:37:02

(in-package :oakcircle-utilities)

;;; General-purpose bit-manipulation macros.

(defmacro mask (bits)
  `(1- (ash 1 (abs ,bits))))

;;; 32-bit-specific types and functions. 

(deftype dword ()
  `(unsigned-byte 32))

(defconstant +dword-mask+ (mask 32))

(declaim (inline mask32 rotl32 +mod32 bitwise-complement-32))

(defun mask32 (n)
  (the dword (logand n +dword-mask+)))

(defun rotl32 (number rotate-bits)
  (let ((rbits (- 32 rotate-bits)))
    (logior (ash (ldb (byte rbits 0) number) rotate-bits)
            (ldb (byte rotate-bits rbits) number))))

(defun +mod32 (&rest params)
  (mask32 (apply #'+ params)))

(defun bitwise-complement-32 (number)
  (mask32 (lognot number)))

(defmacro incf-32 (place &optional (add 1))
  `(setf ,place (mask32 (+ ,place ,add))))

(defun swap-bytes (n bits)
  (ecase bits
         (8 n)
         (16 (rotatef (ldb (byte 8 0) n) (ldb (byte 8 8) n))
             n)
         (32 (rotatef (ldb (byte 8 0) n) (ldb (byte 8 24) n))
             (rotatef (ldb (byte 8 8) n) (ldb (byte 8 16) n))
             n)
         (64 (rotatef (ldb (byte 8  0) n) (ldb (byte 8 56) n))
             (rotatef (ldb (byte 8  8) n) (ldb (byte 8 48) n))
             (rotatef (ldb (byte 8 16) n) (ldb (byte 8 40) n))
             (rotatef (ldb (byte 8 24) n) (ldb (byte 8 32) n))
             n)))

(defun format-hash (a bits-per-item &key lowercase low-endian)
  "General-purpose function for formatting hash values."
  (let ((r ""))
    (dotimes (i (array-dimension a 0))
      (setf r (concatenate 'string r
                           (format nil (format nil "~~~a,'0x"
                                               (/ bits-per-item 4))
                                   (if low-endian
                                       (swap-bytes (aref a i) bits-per-item)
                                       (aref a i))))))
    (if lowercase
        (string-downcase r)
        r)))
