; Copyright (c) 2007 Oak Circle Software, Inc.
; 
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
; 
; The above copyright notice and this permission notice shall be included in
; all copies or substantial portions of the Software.
; 
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
; THE SOFTWARE.

;;;; Created on 2007-10-06 22:43:46

(in-package :oakcircle-utilities)

;;; The md5-specific rounds macros. These aren't perfect examples of macros:
;;; arguments are evaluated multiple times, and not always in the order they're
;;; given. But they abstract out code that has to be repeated several times, so
;;; they do the job they're intended for.

(defmacro md5-do-partial-round (logic a b c d data datarefs s consts)
  `(progn
     (setf ,a (,logic ,a ,b ,c ,d (aref ,data ,(svref datarefs 0)) ,(svref s 0)
                      ,(svref consts 0)))
     (setf ,d (,logic ,d ,a ,b ,c (aref ,data ,(svref datarefs 1)) ,(svref s 1)
                      ,(svref consts 1)))
     (setf ,c (,logic ,c ,d ,a ,b (aref ,data ,(svref datarefs 2)) ,(svref s 2)
                      ,(svref consts 2)))
     (setf ,b (,logic ,b ,c ,d ,a (aref ,data ,(svref datarefs 3)) ,(svref s 3)
                      ,(svref consts 3)))))

(defmacro md5-do-round (logic i data datarefs s consts)
  `(gensym-let ((a (aref ,i 0)) (b (aref ,i 1)) (c (aref ,i 2)) (d (aref ,i 3)))
     (md5-do-partial-round ,logic a b c d ,data ,(subseq datarefs  0  4) ,s
                           ,(subseq consts  0  4))
     (md5-do-partial-round ,logic a b c d ,data ,(subseq datarefs  4  8) ,s
                           ,(subseq consts  4  8))
     (md5-do-partial-round ,logic a b c d ,data ,(subseq datarefs  8 12) ,s
                           ,(subseq consts  8 12))
     (md5-do-partial-round ,logic a b c d ,data ,(subseq datarefs 12 16) ,s
                           ,(subseq consts 12 16))
     (vector a b c d)))

;;; The functions for the transform. I originally implemented these as macros as
;;; well, but changed them because they ran into the inline function limit in
;;; SBCL.

(defun md5-round1 (i data)
  (flet ((doround1 (a b c d x s tx)
           (+mod32 (rotl32 (+ (logior (logand b c)
                                      (logand (bitwise-complement-32 b) d))
                              a x tx) s) b)))
    (declare (inline doround1))
    (md5-do-round doround1 i data
                   #(0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15) #(7 12 17 22)
                   #(#xd76aa478 #xe8c7b756 #x242070db #xc1bdceee #xf57c0faf
                                #x4787c62a #xa8304613 #xfd469501 #x698098d8
                                #x8b44f7af #xffff5bb1 #x895cd7be #x6b901122
                                #xfd987193 #xa679438e #x49b40821))))

(defun md5-round2 (i data)
  (flet ((doround2 (a b c d x s tx)
           (+mod32 (rotl32 (+ (logior (logand b d)
                                      (logand c (bitwise-complement-32 d)))
                              a x tx) s) b)))
    (declare (inline doround2))
    (md5-do-round doround2 i data
                   #(1 6 11 0 5 10 15 4 9 14 3 8 13 2 7 12) #(5 9 14 20)
                   #(#xf61e2562 #xc040b340 #x265e5a51 #xe9b6c7aa #xd62f105d
                                #x02441453 #xd8a1e681 #xe7d3fbc8 #x21e1cde6
                                #xc33707d6 #xf4d50d87 #x455a14ed #xa9e3e905
                                #xfcefa3f8 #x676f02d9 #x8d2a4c8a))))

(defun md5-round3 (i data)
  (flet ((doround3 (a b c d x s tx)
           (+mod32 (rotl32 (+ (logxor b c d) a x tx) s) b)))
    (declare (inline doround3))
    (md5-do-round doround3 i data
                   #(5 8 11 14 1 4 7 10 13 0 3 6 9 12 15 2) #(4 11 16 23)
                   #(#xfffa3942 #x8771f681 #x6d9d6122 #xfde5380c #xa4beea44
                                #x4bdecfa9 #xf6bb4b60 #xbebfbc70 #x289b7ec6
                                #xeaa127fa #xd4ef3085 #x04881d05 #xd9d4d039
                                #xe6db99e5 #x1fa27cf8 #xc4ac5665))))

(defun md5-round4 (i data)
  (flet ((doround4 (a b c d x s tx)
           (+mod32 (rotl32 (+ (logxor c (logior b (bitwise-complement-32 d)))
                              a x tx) s) b)))
    (declare (inline doround4))
    (md5-do-round doround4 i data
                   #(0 7 14 5 12 3 10 1 8 15 6 13 4 11 2 9) #(6 10 15 21)
                   #(#xf4292244 #x432aff97 #xab9423a7 #xfc93a039 #x655b59c3
                                #x8f0ccc92 #xffeff47d #x85845dd1 #x6fa87e4f
                                #xfe2ce6e0 #xa3014314 #x4e0811a1 #xf7537e82
                                #xbd3af235 #x2ad7d2bb #xeb86d391))))

;;; This is the function that ties all of the above together.

(defun md5-transformblock (i data)
  (flet ((byte-array->16-little-endian-dwords (data)
           (flet ((bytes->little-endian-dword (b idx)
                    (let ((r 0))
                      (setf (ldb (byte 8  0) r) (aref b idx)
                            (ldb (byte 8  8) r) (aref b (+ idx 1))
                            (ldb (byte 8 16) r) (aref b (+ idx 2))
                            (ldb (byte 8 24) r) (aref b (+ idx 3)))
                      r)))
             (let ((r (make-array 16 :element-type 'dword)))
               (dotimes (n 16 r)
                 (setf (aref r n) (bytes->little-endian-dword data (* n 4))))))))
    (let* ((x (byte-array->16-little-endian-dwords data))
           (ii (md5-round4
                (md5-round3
                 (md5-round2
                  (md5-round1 i x) x) x) x)))
      (dotimes (n 4 ii)
        (incf-32 (aref ii n) (aref i n))))))

;;; Helper functions used by the md5 function.

(defun make-md5-buffer (&optional original-buffer)
  "Create a new array, of a particular length and type and with a fill-pointer,
   and optionally copy whatever data exists from the original buffer passed into
   this function. Used internally."
  (let ((r (make-array 64 :element-type '(unsigned-byte 8) :fill-pointer 0)))
    (when original-buffer
      (setf (fill-pointer r) (length original-buffer))
      (setf (subseq r 0 (length original-buffer)) original-buffer))
    r))

(defun calculate-md5-of-data (source &optional md5-data)
  "Non-destructively adds more data to a new or existing md5 structure. Used
   internally."
  ;; The flet functions here are mostly to aid in making this self-documenting
  ;; code.
  (flet ((byte-or-character->byte (item)
           (if (characterp item)
                 (char-code item)
                 item))
         (initial-md5-state ()
           (list (vector #x67452301 #xefcdab89 #x98badcfe #x10325476)
                 0 (make-md5-buffer))))
    (declare (inline byte-or-character->byte))
    (when (null md5-data) (setf md5-data (initial-md5-state)))
    (let ((hashvalue (first md5-data))
          (bytesprocessed (second md5-data))
          (datablock (make-md5-buffer (third md5-data))))
      (let ((end-source-index (length source))
            (source-index 0))
        (loop while (< source-index end-source-index) do
              (loop while (and (< source-index end-source-index)
                               (< (fill-pointer datablock) 64)) do
                    (vector-push (byte-or-character->byte (aref source source-index))
                                 datablock)
                    (incf source-index))
              (when (= (fill-pointer datablock) 64)
                (setf hashvalue (md5-transformblock hashvalue datablock))
                (setf (fill-pointer datablock) 0)
                (incf bytesprocessed 64))))
      (list hashvalue bytesprocessed datablock))))

(defun finish-md5 (md5-data)
  "Non-destructively adds the size-padding and calculates the end of the hash.
   Used internally."
  (flet ((pad-with-zeros (a to)
           (let ((old-length (fill-pointer a)))
             (setf (fill-pointer a) to)
             (fill a 0 :start old-length :end to))))
    (let ((hashvalue (first md5-data))
          (bytesprocessed (second md5-data))
          (datablock (make-md5-buffer (third md5-data))))
      (incf bytesprocessed (fill-pointer datablock))
      (let ((bitsprocessed (* bytesprocessed 8)))
        (vector-push #x80 datablock)
        (when (> (fill-pointer datablock) 56)
          (pad-with-zeros datablock 64)
          (setf hashvalue (md5-transformblock hashvalue datablock))
          (setf (fill-pointer datablock) 0))
        (pad-with-zeros datablock 56)
        (dotimes (i 8)
          (vector-push (ldb (byte 8 (* i 8)) bitsprocessed) datablock))
        (setf hashvalue (md5-transformblock hashvalue datablock))
        (setf (fill-pointer datablock) 0))
      hashvalue)))

;;; The externally-visible md5 function. 

(defun md5 (source &key add-to partial)
  "Use this function to create an MD5 hash out of a single block of data (an
   ASCII or UTF-8 string, or an array of unsigned-byte 8s) by passing it the
   data to hash. If you don't have all of the data you want to hash as a single
   block, then specify the ':partial t' keyword parameter and it will return the
   internal md5 data instead, which you can pass to the next md5 call using the
   :add-to keyword parameter."
  (if (eq source :test)
      (test-md5)
      (let ((hashvalue (calculate-md5-of-data source add-to)))
        (if partial
            hashvalue
            (format-hash (finish-md5 hashvalue) 32
                         :lowercase t :low-endian t)))))

;;; The test function.

(defun test-md5 ()
  "Tests the md5 functions. Only two tests are actually needed: one with 56 or
   fewer bytes, and one with 57 or more. The rest of the tests are to confirm
   the operation of certain features. Returns T if all is well, NIL if there's
   a problem -- try (trace md5) before running it to see where the problem is."
  (let ((abc-string "abc")
        (abc-array (vector (char-code #\a) (char-code #\b) (char-code #\c)))
        (abc-string-hashvalue "900150983cd24fb0d6963f7d28e17f72"))
    (let ((string56 "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDEFGHIJ")
          (string56-hashvalue "2402a47bfc72cabd03386832132874f2")
          (string56+16 "KLMNOPQRSTUVWXYZ"))
      (let ((string72 (concatenate 'string string56 string56+16))
            (string72-hashvalue "d3ee71770535ca94a5a762650ba49d60"))
        (and (string= (md5 string56) string56-hashvalue)
             (string= (md5 string72) string72-hashvalue)
             (string= (md5 string56+16 :add-to (md5 string56 :partial t))
                      string72-hashvalue)
             (string= (md5 abc-string) abc-string-hashvalue)
             (string= (md5 abc-array)  abc-string-hashvalue))))))
