; Copyright (c) 2007 Oak Circle Software, Inc.
; 
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
; 
; The above copyright notice and this permission notice shall be included in
; all copies or substantial portions of the Software.
; 
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
; THE SOFTWARE.

;;;; 2007-09-05 16:26:47

(in-package :oakcircle-utilities)

(defmacro gensym-let (vars-and-values &body body)
  "Creates a gensym in place of each variable listed in the vars-and-values
  list, replacing all instances of that variable in the body with the generated
  symbol, and optionally assigning it a value. Meant for use in other macros.
  The syntax is identical to that of let."
  (labels ((first-or-only (x)
             (if (listp x)
                 (if (<= 1 (length x) 2)
                     (the symbol (first x))
                     (error "The GENSYM-LET binding spec ~a is malformed." x))
                 (the symbol x)))
           (second-or-nil (x) (if (listp x) (second x) nil)))
    (let ((newlets nil))
      (dolist (i vars-and-values `(let ,newlets ,@body))
        (let ((g (gensym (concatenate 'string (symbol-name (first-or-only i)) "-"))))
          (unless (eql (second-or-nil i) nil)
            (setf newlets (append newlets `((,g ,(second-or-nil i))))))
          (setf body (subst g (first-or-only i) body)))))))

(defmacro do-while (test &body body)
  `(do ()
       ((not ,test))
     ,@body))

(defmacro do-until (test &body body)
  `(do ()
       ((,test))
     ,@body))

(defmacro do-forever (&body body)
  `(do ()
       ((nil))
     ,@body))
